import flags
import numpy
import pydrake.common
from typing import Any, Callable, ClassVar, Iterator, overload

from_sympy: Callable
logical_and: Callable
logical_or: Callable
to_sympy: Callable

class Expression:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, constant: float) -> None: ...
    @overload
    def __init__(self, var: Variable) -> None: ...
    def Differentiate(self, x: Variable) -> Expression: ...
    def EqualTo(self, arg0: Expression) -> bool: ...
    @overload
    def Evaluate(self, env: dict[Variable, float] = ..., generator: pydrake.common.RandomGenerator = ...) -> float: ...
    @overload
    def Evaluate(self, generator: pydrake.common.RandomGenerator) -> float: ...
    def EvaluatePartial(self, env: dict[Variable, float]) -> Expression: ...
    def Expand(self) -> Expression: ...
    def GetFreeVariables(self) -> Variables: ...
    def GetVariables(self) -> Variables: ...
    def Jacobian(self, vars: numpy.ndarray[object[m, 1]]) -> numpy.ndarray[object[1, n]]: ...
    @overload
    def Substitute(self, var: Variable, e: Expression) -> Expression: ...
    @overload
    def Substitute(self, s: dict[Variable, Expression]) -> Expression: ...
    def Unapply(self) -> object: ...
    def abs(self) -> Expression: ...
    def acos(self) -> Expression: ...
    def arccos(self) -> Expression: ...
    def arcsin(self) -> Expression: ...
    def arctan(self) -> Expression: ...
    @overload
    def arctan2(self, x: Expression) -> Expression: ...
    @overload
    def arctan2(self, y, x) -> Any: ...
    def asin(self) -> Expression: ...
    def atan(self) -> Expression: ...
    @overload
    def atan2(self, x: Expression) -> Expression: ...
    @overload
    def atan2(self, y, x) -> Any: ...
    def ceil(self) -> Expression: ...
    def cos(self) -> Expression: ...
    def cosh(self) -> Expression: ...
    def exp(self) -> Expression: ...
    def floor(self) -> Expression: ...
    def get_kind(self) -> ExpressionKind: ...
    def is_polynomial(self) -> bool: ...
    def log(self) -> Expression: ...
    def max(self, arg0: Expression) -> Expression: ...
    def min(self, arg0: Expression) -> Expression: ...
    @overload
    def pow(self, arg0: float) -> Expression: ...
    @overload
    def pow(self, arg0: Expression) -> Expression: ...
    def sin(self) -> Expression: ...
    def sinh(self) -> Expression: ...
    def sqrt(self) -> Expression: ...
    def tan(self) -> Expression: ...
    def tanh(self) -> Expression: ...
    def to_string(self) -> str: ...
    def __abs__(self) -> Expression: ...
    @overload
    def __add__(self, arg0: Expression) -> Expression: ...
    @overload
    def __add__(self, arg0: Variable) -> Expression: ...
    @overload
    def __add__(self, arg0: float) -> Expression: ...
    def __ceil__(self) -> Expression: ...
    @overload
    def __copy__(self) -> Expression: ...
    @overload
    def __copy__(self) -> Expression: ...
    def __deepcopy__(self, arg0: dict) -> Expression: ...
    @overload
    def __eq__(self, arg0: Expression) -> Formula: ...
    @overload
    def __eq__(self, arg0: Variable) -> Formula: ...
    @overload
    def __eq__(self, arg0: float) -> Formula: ...
    def __floor__(self) -> Expression: ...
    @overload
    def __ge__(self, arg0: Expression) -> Formula: ...
    @overload
    def __ge__(self, arg0: Variable) -> Formula: ...
    @overload
    def __ge__(self, arg0: float) -> Formula: ...
    @overload
    def __gt__(self, arg0: Expression) -> Formula: ...
    @overload
    def __gt__(self, arg0: Variable) -> Formula: ...
    @overload
    def __gt__(self, arg0: float) -> Formula: ...
    @overload
    def __iadd__(self, arg0: Expression) -> Expression: ...
    @overload
    def __iadd__(self, arg0: Variable) -> Expression: ...
    @overload
    def __iadd__(self, arg0: float) -> Expression: ...
    @overload
    def __imul__(self, arg0: Expression) -> Expression: ...
    @overload
    def __imul__(self, arg0: Variable) -> Expression: ...
    @overload
    def __imul__(self, arg0: float) -> Expression: ...
    @overload
    def __isub__(self, arg0: Expression) -> Expression: ...
    @overload
    def __isub__(self, arg0: Variable) -> Expression: ...
    @overload
    def __isub__(self, arg0: float) -> Expression: ...
    @overload
    def __itruediv__(self, arg0: Expression) -> Expression: ...
    @overload
    def __itruediv__(self, arg0: Variable) -> Expression: ...
    @overload
    def __itruediv__(self, arg0: float) -> Expression: ...
    @overload
    def __le__(self, arg0: Expression) -> Formula: ...
    @overload
    def __le__(self, arg0: Variable) -> Formula: ...
    @overload
    def __le__(self, arg0: float) -> Formula: ...
    @overload
    def __lt__(self, arg0: Expression) -> Formula: ...
    @overload
    def __lt__(self, arg0: Variable) -> Formula: ...
    @overload
    def __lt__(self, arg0: float) -> Formula: ...
    @overload
    def __mul__(self, arg0: Expression) -> Expression: ...
    @overload
    def __mul__(self, arg0: Variable) -> Expression: ...
    @overload
    def __mul__(self, arg0: float) -> Expression: ...
    @overload
    def __ne__(self, arg0: Expression) -> Formula: ...
    @overload
    def __ne__(self, arg0: Variable) -> Formula: ...
    @overload
    def __ne__(self, arg0: float) -> Formula: ...
    def __neg__(self) -> Expression: ...
    def __pos__(self) -> Expression: ...
    @overload
    def __pow__(self, arg0: float) -> Expression: ...
    @overload
    def __pow__(self, arg0: Variable) -> Expression: ...
    @overload
    def __pow__(self, arg0: Expression) -> Expression: ...
    @overload
    def __radd__(self, arg0: Variable) -> Expression: ...
    @overload
    def __radd__(self, arg0: float) -> Expression: ...
    @overload
    def __rmul__(self, arg0: Variable) -> Expression: ...
    @overload
    def __rmul__(self, arg0: float) -> Expression: ...
    @overload
    def __rsub__(self, arg0: Variable) -> Expression: ...
    @overload
    def __rsub__(self, arg0: float) -> Expression: ...
    @overload
    def __rtruediv__(self, arg0: Variable) -> Expression: ...
    @overload
    def __rtruediv__(self, arg0: float) -> Expression: ...
    @overload
    def __sub__(self, arg0: Expression) -> Expression: ...
    @overload
    def __sub__(self, arg0: Variable) -> Expression: ...
    @overload
    def __sub__(self, arg0: float) -> Expression: ...
    @overload
    def __truediv__(self, arg0: Expression) -> Expression: ...
    @overload
    def __truediv__(self, arg0: Variable) -> Expression: ...
    @overload
    def __truediv__(self, arg0: float) -> Expression: ...

class ExpressionKind:
    __members__: ClassVar[dict] = ...  # read-only
    Abs: ClassVar[ExpressionKind] = ...
    Acos: ClassVar[ExpressionKind] = ...
    Add: ClassVar[ExpressionKind] = ...
    Asin: ClassVar[ExpressionKind] = ...
    Atan: ClassVar[ExpressionKind] = ...
    Atan2: ClassVar[ExpressionKind] = ...
    Ceil: ClassVar[ExpressionKind] = ...
    Constant: ClassVar[ExpressionKind] = ...
    Cos: ClassVar[ExpressionKind] = ...
    Cosh: ClassVar[ExpressionKind] = ...
    Div: ClassVar[ExpressionKind] = ...
    Exp: ClassVar[ExpressionKind] = ...
    Floor: ClassVar[ExpressionKind] = ...
    IfThenElse: ClassVar[ExpressionKind] = ...
    Log: ClassVar[ExpressionKind] = ...
    Max: ClassVar[ExpressionKind] = ...
    Min: ClassVar[ExpressionKind] = ...
    Mul: ClassVar[ExpressionKind] = ...
    NaN: ClassVar[ExpressionKind] = ...
    Pow: ClassVar[ExpressionKind] = ...
    Sin: ClassVar[ExpressionKind] = ...
    Sinh: ClassVar[ExpressionKind] = ...
    Sqrt: ClassVar[ExpressionKind] = ...
    Tan: ClassVar[ExpressionKind] = ...
    Tanh: ClassVar[ExpressionKind] = ...
    UninterpretedFunction: ClassVar[ExpressionKind] = ...
    Var: ClassVar[ExpressionKind] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class Formula:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, value: bool) -> None: ...
    @overload
    def __init__(self, var: Variable) -> None: ...
    def EqualTo(self, arg0: Formula) -> bool: ...
    def Evaluate(self, env: dict[Variable, float] = ...) -> bool: ...
    @staticmethod
    def False_() -> Formula: ...
    def GetFreeVariables(self) -> Variables: ...
    @overload
    def Substitute(self, var: Variable, e: Expression) -> Formula: ...
    @overload
    def Substitute(self, var: Variable, e: Variable) -> Formula: ...
    @overload
    def Substitute(self, var: Variable, e: float) -> Formula: ...
    @overload
    def Substitute(self, s: dict[Variable, Expression]) -> Formula: ...
    @staticmethod
    def True_() -> Formula: ...
    def Unapply(self) -> object: ...
    def get_kind(self) -> FormulaKind: ...
    def to_string(self) -> str: ...
    def __bool__(self) -> bool: ...
    def __eq__(self, arg0: Formula) -> bool: ...
    def __hash__(self) -> int: ...
    def __ne__(self, arg0: Formula) -> bool: ...
    def __nonzero__(self) -> None: ...

class FormulaKind:
    __members__: ClassVar[dict] = ...  # read-only
    And: ClassVar[FormulaKind] = ...
    Eq: ClassVar[FormulaKind] = ...
    False_: ClassVar[FormulaKind] = ...
    Forall: ClassVar[FormulaKind] = ...
    Geq: ClassVar[FormulaKind] = ...
    Gt: ClassVar[FormulaKind] = ...
    Isnan: ClassVar[FormulaKind] = ...
    Leq: ClassVar[FormulaKind] = ...
    Lt: ClassVar[FormulaKind] = ...
    Neq: ClassVar[FormulaKind] = ...
    Not: ClassVar[FormulaKind] = ...
    Or: ClassVar[FormulaKind] = ...
    PositiveSemidefinite: ClassVar[FormulaKind] = ...
    True_: ClassVar[FormulaKind] = ...
    Var: ClassVar[FormulaKind] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class Monomial:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, var: Variable) -> None: ...
    @overload
    def __init__(self, var: Variable, exponent: int) -> None: ...
    @overload
    def __init__(self, powers: dict[Variable, int]) -> None: ...
    @overload
    def __init__(self, vars: numpy.ndarray[object[m, 1]], exponents: numpy.ndarray[numpy.int32[m, 1]]) -> None: ...
    def EqualTo(self, arg0: Monomial) -> bool: ...
    @overload
    def Evaluate(self, env: dict[Variable, float]) -> float: ...
    @overload
    def Evaluate(self, vars: numpy.ndarray[object[m, 1]], vars_values: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    def EvaluatePartial(self, env: dict[Variable, float]) -> tuple[float, Monomial]: ...
    def GetVariables(self) -> Variables: ...
    def ToExpression(self) -> Expression: ...
    def degree(self, v: Variable) -> int: ...
    def get_powers(self) -> dict[Variable, int]: ...
    def pow_in_place(self, p: int) -> Monomial: ...
    def total_degree(self) -> int: ...
    def __add__(self, arg0: Expression) -> Expression: ...
    def __eq__(self, arg0: Monomial) -> bool: ...
    def __hash__(self) -> int: ...
    def __imul__(self, arg0: Monomial) -> Monomial: ...
    @overload
    def __mul__(self, arg0: Monomial) -> Monomial: ...
    @overload
    def __mul__(self, arg0: float) -> Polynomial: ...
    @overload
    def __mul__(self, arg0: Expression) -> Expression: ...
    @overload
    def __ne__(self, arg0: Monomial) -> bool: ...
    @overload
    def __ne__(self, arg0: Monomial) -> bool: ...
    def __pow__(self, arg0: int) -> Monomial: ...
    def __radd__(self, arg0: Expression) -> Expression: ...
    @overload
    def __rmul__(self, arg0: float) -> Polynomial: ...
    @overload
    def __rmul__(self, arg0: Expression) -> Expression: ...
    def __rsub__(self, arg0: Expression) -> Expression: ...
    def __rtruediv__(self, arg0: Expression) -> Expression: ...
    def __sub__(self, arg0: Expression) -> Expression: ...
    def __truediv__(self, arg0: Expression) -> Expression: ...

class Polynomial:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, map: dict[Monomial, Expression]) -> None: ...
    @overload
    def __init__(self, m: Monomial) -> None: ...
    @overload
    def __init__(self, e: Expression) -> None: ...
    @overload
    def __init__(self, e: Expression, indeterminates: Variables) -> None: ...
    @overload
    def __init__(self, e: Expression, indeterminates: numpy.ndarray[object[m, 1]]) -> None: ...
    def AddProduct(self, coeff: Expression, m: Monomial) -> Polynomial: ...
    def CoefficientsAlmostEqual(self, p: Polynomial, tolerance: float) -> bool: ...
    def Degree(self, v: Variable) -> int: ...
    def Differentiate(self, x: Variable) -> Polynomial: ...
    def EqualTo(self, arg0: Polynomial) -> bool: ...
    def Evaluate(self, env: dict[Variable, float]) -> float: ...
    def EvaluateIndeterminates(self, indeterminates: numpy.ndarray[object[m, 1]], indeterminates_values: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    @overload
    def EvaluatePartial(self, env: dict[Variable, float]) -> Polynomial: ...
    @overload
    def EvaluatePartial(self, var: Variable, c: float) -> Polynomial: ...
    def EvaluateWithAffineCoefficients(self, indeterminates: numpy.ndarray[object[m, 1]], indeterminates_values: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> tuple[numpy.ndarray[numpy.float64[m, n]], numpy.ndarray[object[m, 1]], numpy.ndarray[numpy.float64[m, 1]]]: ...
    def Expand(self) -> Polynomial: ...
    @overload
    def Integrate(self, x: Variable) -> Polynomial: ...
    @overload
    def Integrate(self, x: Variable, a: float, b: float) -> Polynomial: ...
    def IsEven(self) -> bool: ...
    def IsOdd(self) -> bool: ...
    def Jacobian(self, vars: numpy.ndarray[object[m, 1]]) -> numpy.ndarray[object[1, n]]: ...
    def RemoveTermsWithSmallCoefficients(self, coefficient_tol: float) -> Polynomial: ...
    def Roots(self) -> numpy.ndarray[numpy.complex128[m, 1]]: ...
    def SetIndeterminates(self, new_indeterminates: Variables) -> None: ...
    def SubstituteAndExpand(self, indeterminate_substitution: dict[Variable, Polynomial], substitutions_cached_data: SubstituteAndExpandCacheData = ...) -> Polynomial: ...
    def ToExpression(self) -> Expression: ...
    def TotalDegree(self) -> int: ...
    def decision_variables(self) -> Variables: ...
    def indeterminates(self) -> Variables: ...
    @overload
    def monomial_to_coefficient_map(self) -> dict[Monomial, Expression]: ...
    @overload
    def monomial_to_coefficient_map(self) -> Any: ...
    @overload
    def __add__(self, arg0: Polynomial) -> Polynomial: ...
    @overload
    def __add__(self, arg0: Monomial) -> Polynomial: ...
    @overload
    def __add__(self, arg0: float) -> Polynomial: ...
    @overload
    def __add__(self, arg0: Variable) -> Polynomial: ...
    @overload
    def __add__(self, arg0: Expression) -> Expression: ...
    def __eq__(self, arg0: Polynomial) -> Formula: ...
    def __hash__(self) -> int: ...
    @overload
    def __mul__(self, arg0: Polynomial) -> Polynomial: ...
    @overload
    def __mul__(self, arg0: Monomial) -> Polynomial: ...
    @overload
    def __mul__(self, arg0: float) -> Polynomial: ...
    @overload
    def __mul__(self, arg0: Variable) -> Polynomial: ...
    @overload
    def __mul__(self, arg0: Expression) -> Expression: ...
    def __ne__(self, arg0: Polynomial) -> Formula: ...
    def __neg__(self) -> Polynomial: ...
    def __pow__(self, arg0: int) -> Polynomial: ...
    @overload
    def __radd__(self, arg0: Monomial) -> Polynomial: ...
    @overload
    def __radd__(self, arg0: float) -> Polynomial: ...
    @overload
    def __radd__(self, arg0: Variable) -> Polynomial: ...
    @overload
    def __radd__(self, arg0: Expression) -> Expression: ...
    @overload
    def __rmul__(self, arg0: Monomial) -> Polynomial: ...
    @overload
    def __rmul__(self, arg0: float) -> Polynomial: ...
    @overload
    def __rmul__(self, arg0: Variable) -> Polynomial: ...
    @overload
    def __rmul__(self, arg0: Expression) -> Expression: ...
    @overload
    def __rsub__(self, arg0: Monomial) -> Polynomial: ...
    @overload
    def __rsub__(self, arg0: float) -> Polynomial: ...
    @overload
    def __rsub__(self, arg0: Variable) -> Polynomial: ...
    @overload
    def __rsub__(self, arg0: Expression) -> Expression: ...
    @overload
    def __rtruediv__(self, arg0: float) -> Expression: ...
    @overload
    def __rtruediv__(self, arg0: Expression) -> Expression: ...
    @overload
    def __sub__(self, arg0: Polynomial) -> Polynomial: ...
    @overload
    def __sub__(self, arg0: Monomial) -> Polynomial: ...
    @overload
    def __sub__(self, arg0: float) -> Polynomial: ...
    @overload
    def __sub__(self, arg0: Variable) -> Polynomial: ...
    @overload
    def __sub__(self, arg0: Expression) -> Expression: ...
    @overload
    def __truediv__(self, arg0: float) -> Polynomial: ...
    @overload
    def __truediv__(self, arg0: Expression) -> Expression: ...

class RationalFunction:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, numerator: Polynomial, denominator: Polynomial) -> None: ...
    @overload
    def __init__(self, p: Polynomial) -> None: ...
    @overload
    def __init__(self, m: Monomial) -> None: ...
    @overload
    def __init__(self, c: float) -> None: ...
    @overload
    def __init__(self) -> None: ...
    def EqualTo(self, f: RationalFunction) -> bool: ...
    def Evaluate(self, env: dict[Variable, float]) -> float: ...
    def SetIndeterminates(self, new_indeterminates: Variables) -> None: ...
    def ToExpression(self) -> Expression: ...
    def denominator(self) -> Polynomial: ...
    def numerator(self) -> Polynomial: ...
    @overload
    def __add__(self, arg0: RationalFunction) -> RationalFunction: ...
    @overload
    def __add__(self, arg0: float) -> RationalFunction: ...
    @overload
    def __add__(self, arg0: Polynomial) -> RationalFunction: ...
    @overload
    def __add__(self, arg0: Monomial) -> RationalFunction: ...
    def __eq__(self, arg0: RationalFunction) -> Formula: ...
    @overload
    def __mul__(self, arg0: RationalFunction) -> RationalFunction: ...
    @overload
    def __mul__(self, arg0: float) -> RationalFunction: ...
    @overload
    def __mul__(self, arg0: Polynomial) -> RationalFunction: ...
    @overload
    def __mul__(self, arg0: Monomial) -> RationalFunction: ...
    def __ne__(self, arg0: RationalFunction) -> Formula: ...
    def __neg__(self) -> RationalFunction: ...
    @overload
    def __radd__(self, arg0: float) -> RationalFunction: ...
    @overload
    def __radd__(self, arg0: Polynomial) -> RationalFunction: ...
    @overload
    def __radd__(self, arg0: Monomial) -> RationalFunction: ...
    @overload
    def __rmul__(self, arg0: float) -> RationalFunction: ...
    @overload
    def __rmul__(self, arg0: Polynomial) -> RationalFunction: ...
    @overload
    def __rmul__(self, arg0: Monomial) -> RationalFunction: ...
    @overload
    def __rsub__(self, arg0: float) -> RationalFunction: ...
    @overload
    def __rsub__(self, arg0: Polynomial) -> RationalFunction: ...
    @overload
    def __rsub__(self, arg0: Monomial) -> RationalFunction: ...
    @overload
    def __rtruediv__(self, arg0: float) -> RationalFunction: ...
    @overload
    def __rtruediv__(self, arg0: Polynomial) -> RationalFunction: ...
    @overload
    def __rtruediv__(self, arg0: Monomial) -> RationalFunction: ...
    @overload
    def __sub__(self, arg0: RationalFunction) -> RationalFunction: ...
    @overload
    def __sub__(self, arg0: float) -> RationalFunction: ...
    @overload
    def __sub__(self, arg0: Polynomial) -> RationalFunction: ...
    @overload
    def __sub__(self, arg0: Monomial) -> RationalFunction: ...
    @overload
    def __truediv__(self, arg0: RationalFunction) -> RationalFunction: ...
    @overload
    def __truediv__(self, arg0: float) -> RationalFunction: ...
    @overload
    def __truediv__(self, arg0: Polynomial) -> RationalFunction: ...
    @overload
    def __truediv__(self, arg0: Monomial) -> RationalFunction: ...

class SinCos:
    c: Variable
    s: Variable
    type: SinCosSubstitutionType
    def __init__(self, s: Variable, c: Variable, type: SinCosSubstitutionType = ...) -> None: ...

class SinCosSubstitutionType:
    __members__: ClassVar[dict] = ...  # read-only
    __entries: ClassVar[dict] = ...
    kAngle: ClassVar[SinCosSubstitutionType] = ...
    kHalfAnglePreferCos: ClassVar[SinCosSubstitutionType] = ...
    kHalfAnglePreferSin: ClassVar[SinCosSubstitutionType] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class SubstituteAndExpandCacheData:
    def __init__(self) -> None: ...
    def get_data(self) -> dict[Monomial, Polynomial]: ...

class Variable:
    class Type:
        __members__: ClassVar[dict] = ...  # read-only
        BINARY: ClassVar[Variable.Type] = ...
        BOOLEAN: ClassVar[Variable.Type] = ...
        CONTINUOUS: ClassVar[Variable.Type] = ...
        INTEGER: ClassVar[Variable.Type] = ...
        RANDOM_EXPONENTIAL: ClassVar[Variable.Type] = ...
        RANDOM_GAUSSIAN: ClassVar[Variable.Type] = ...
        RANDOM_UNIFORM: ClassVar[Variable.Type] = ...
        __entries: ClassVar[dict] = ...
        def __init__(self, value: int) -> None: ...
        def __eq__(self, other: object) -> bool: ...
        def __hash__(self) -> int: ...
        def __index__(self) -> int: ...
        def __int__(self) -> int: ...
        def __ne__(self, other: object) -> bool: ...
        @property
        def name(self) -> str: ...
        @property
        def value(self) -> int: ...
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, name: str, type: Variable.Type = ...) -> None: ...
    def EqualTo(self, arg0: Variable) -> bool: ...
    def abs(self) -> Expression: ...
    def acos(self) -> Expression: ...
    def arccos(self) -> Expression: ...
    def arcsin(self) -> Expression: ...
    def arctan(self) -> Expression: ...
    @overload
    def arctan2(self, x: Variable) -> Expression: ...
    @overload
    def arctan2(self, y, x) -> Any: ...
    def asin(self) -> Expression: ...
    def atan(self) -> Expression: ...
    @overload
    def atan2(self, x: Variable) -> Expression: ...
    @overload
    def atan2(self, y, x) -> Any: ...
    def ceil(self) -> Expression: ...
    def cos(self) -> Expression: ...
    def cosh(self) -> Expression: ...
    def exp(self) -> Expression: ...
    def floor(self) -> Expression: ...
    def get_id(self) -> int: ...
    def get_name(self) -> str: ...
    def get_type(self) -> Variable.Type: ...
    def is_dummy(self) -> bool: ...
    def log(self) -> Expression: ...
    def max(self, arg0: Variable) -> Expression: ...
    def min(self, arg0: Variable) -> Expression: ...
    @overload
    def pow(self, arg0: float) -> Expression: ...
    @overload
    def pow(self, arg0: Expression) -> Expression: ...
    def sin(self) -> Expression: ...
    def sinh(self) -> Expression: ...
    def sqrt(self) -> Expression: ...
    def tan(self) -> Expression: ...
    def tanh(self) -> Expression: ...
    def __abs__(self) -> Expression: ...
    @overload
    def __add__(self, arg0: Variable) -> Expression: ...
    @overload
    def __add__(self, arg0: float) -> Expression: ...
    def __ceil__(self) -> Expression: ...
    def __copy__(self) -> Variable: ...
    def __deepcopy__(self, arg0: dict) -> Variable: ...
    @overload
    def __eq__(self, arg0: Expression) -> Formula: ...
    @overload
    def __eq__(self, arg0: Variable) -> Formula: ...
    @overload
    def __eq__(self, arg0: float) -> Formula: ...
    def __floor__(self) -> Expression: ...
    @overload
    def __ge__(self, arg0: Expression) -> Formula: ...
    @overload
    def __ge__(self, arg0: Variable) -> Formula: ...
    @overload
    def __ge__(self, arg0: float) -> Formula: ...
    @overload
    def __gt__(self, arg0: Expression) -> Formula: ...
    @overload
    def __gt__(self, arg0: Variable) -> Formula: ...
    @overload
    def __gt__(self, arg0: float) -> Formula: ...
    def __hash__(self) -> int: ...
    @overload
    def __le__(self, arg0: Expression) -> Formula: ...
    @overload
    def __le__(self, arg0: Variable) -> Formula: ...
    @overload
    def __le__(self, arg0: float) -> Formula: ...
    @overload
    def __lt__(self, arg0: Expression) -> Formula: ...
    @overload
    def __lt__(self, arg0: Variable) -> Formula: ...
    @overload
    def __lt__(self, arg0: float) -> Formula: ...
    @overload
    def __mul__(self, arg0: Variable) -> Expression: ...
    @overload
    def __mul__(self, arg0: float) -> Expression: ...
    @overload
    def __ne__(self, arg0: Expression) -> Formula: ...
    @overload
    def __ne__(self, arg0: Variable) -> Formula: ...
    @overload
    def __ne__(self, arg0: float) -> Formula: ...
    def __neg__(self) -> Expression: ...
    def __pos__(self) -> Expression: ...
    @overload
    def __pow__(self, arg0: float) -> Expression: ...
    @overload
    def __pow__(self, arg0: Variable) -> Expression: ...
    @overload
    def __pow__(self, arg0: Expression) -> Expression: ...
    def __radd__(self, arg0: float) -> Expression: ...
    def __rmul__(self, arg0: float) -> Expression: ...
    def __rsub__(self, arg0: float) -> Expression: ...
    def __rtruediv__(self, arg0: float) -> Expression: ...
    @overload
    def __sub__(self, arg0: Variable) -> Expression: ...
    @overload
    def __sub__(self, arg0: float) -> Expression: ...
    @overload
    def __truediv__(self, arg0: Variable) -> Expression: ...
    @overload
    def __truediv__(self, arg0: float) -> Expression: ...

class Variables:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, arg0: numpy.ndarray[object[m, 1]]) -> None: ...
    def EqualTo(self, arg0: Variables) -> bool: ...
    def IsStrictSubsetOf(self, vars: Variables) -> bool: ...
    def IsStrictSupersetOf(self, vars: Variables) -> bool: ...
    def IsSubsetOf(self, vars: Variables) -> bool: ...
    def IsSupersetOf(self, vars: Variables) -> bool: ...
    def empty(self) -> bool: ...
    @overload
    def erase(self, key: Variable) -> int: ...
    @overload
    def erase(self, vars: Variables) -> int: ...
    def include(self, key: Variable) -> bool: ...
    @overload
    def insert(self, var: Variable) -> None: ...
    @overload
    def insert(self, vars: Variables) -> None: ...
    def size(self) -> int: ...
    def to_string(self) -> str: ...
    @overload
    def __add__(self, arg0: Variables) -> Variables: ...
    @overload
    def __add__(self, arg0: Variable) -> Variables: ...
    def __contains__(self, arg0: Variable) -> bool: ...
    def __eq__(self, arg0: Variables) -> bool: ...
    def __hash__(self) -> int: ...
    def __iter__(self) -> Iterator[Variable]: ...
    def __len__(self) -> int: ...
    def __lt__(self, arg0: Variables) -> bool: ...
    def __radd__(self, arg0: Variable) -> Variables: ...
    @overload
    def __sub__(self, arg0: Variables) -> Variables: ...
    @overload
    def __sub__(self, arg0: Variable) -> Variables: ...

def CalcMonomialBasisOrderUpToOne(x: Variables, sort_monomial: bool = ...) -> numpy.ndarray[object[m, 1]]: ...
@overload
def CalcPolynomialWLowerTriangularPart(monomial_basis: numpy.ndarray[object[m, 1]], gram_lower: numpy.ndarray[numpy.float64[m, 1]]) -> Polynomial: ...
@overload
def CalcPolynomialWLowerTriangularPart(monomial_basis: numpy.ndarray[object[m, 1]], gram_lower: numpy.ndarray[object[m, 1]]) -> Polynomial: ...
@overload
def CalcPolynomialWLowerTriangularPart(monomial_basis: numpy.ndarray[object[m, 1]], gram_lower: numpy.ndarray[object[m, 1]]) -> Polynomial: ...
def DecomposeAffineExpression(e: Expression, map_var_to_index: dict[int, int]) -> tuple[numpy.ndarray[numpy.float64[1, n]], float]: ...
@overload
def DecomposeAffineExpressions(expressions: numpy.ndarray[object[m, 1]], vars: numpy.ndarray[object[m, 1]]) -> tuple[numpy.ndarray[numpy.float64[m, n]], numpy.ndarray[numpy.float64[m, 1]]]: ...
@overload
def DecomposeAffineExpressions(v: numpy.ndarray[object[m, 1]]) -> tuple[numpy.ndarray[numpy.float64[m, n]], numpy.ndarray[numpy.float64[m, 1]], numpy.ndarray[object[m, 1]]]: ...
def DecomposeL2NormExpression(e: Expression, psd_tol: float = ..., coefficient_tol: float = ...) -> tuple[bool, numpy.ndarray[numpy.float64[m, n]], numpy.ndarray[numpy.float64[m, 1]], numpy.ndarray[object[m, 1]]]: ...
def DecomposeLinearExpressions(expressions: numpy.ndarray[object[m, 1]], vars: numpy.ndarray[object[m, 1]]) -> numpy.ndarray[numpy.float64[m, n]]: ...
def DecomposeLumpedParameters(f: numpy.ndarray[object[m, 1]], parameters: numpy.ndarray[object[m, 1]]) -> tuple[numpy.ndarray[object[m, n]], numpy.ndarray[object[m, 1]], numpy.ndarray[object[m, 1]]]: ...
def DecomposeQuadraticPolynomial(poly: Polynomial, map_var_to_index: dict[int, int]) -> tuple[numpy.ndarray[numpy.float64[m, n]], numpy.ndarray[numpy.float64[m, 1]], float]: ...
@overload
def Evaluate(m: numpy.ndarray[object[m, n]], env: dict[Variable, float] = ..., generator: pydrake.common.RandomGenerator = ...) -> numpy.ndarray[numpy.float64[m, n]]: ...
@overload
def Evaluate(m: numpy.ndarray[object[m, n]], env: dict[Variable, float]) -> numpy.ndarray[numpy.float64[m, n]]: ...
def EvenDegreeMonomialBasis(vars: Variables, degree: int) -> numpy.ndarray[object[m, 1]]: ...
@overload
def ExtractVariablesFromExpression(e: Expression) -> tuple[numpy.ndarray[object[m, 1]], dict[int, int]]: ...
@overload
def ExtractVariablesFromExpression(expressions: numpy.ndarray[object[m, 1]]) -> tuple[numpy.ndarray[object[m, 1]], dict[int, int]]: ...
def GetVariableVector(expressions: numpy.ndarray[object[m, 1]]) -> numpy.ndarray[object[m, 1]]: ...
@overload
def IsAffine(m: numpy.ndarray[object[m, n], flags.f_contiguous], vars: Variables) -> bool: ...
@overload
def IsAffine(m: numpy.ndarray[object[m, n], flags.f_contiguous]) -> bool: ...
@overload
def Jacobian(f: numpy.ndarray[object[m, 1]], vars: numpy.ndarray[object[m, 1]]) -> numpy.ndarray[object[m, n]]: ...
@overload
def Jacobian(f: numpy.ndarray[object[m, 1]], vars: numpy.ndarray[object[m, 1]]) -> numpy.ndarray[object[m, n]]: ...
def MakeMatrixBinaryVariable(rows: int, cols: int, name: str) -> numpy.ndarray[object[m, n]]: ...
def MakeMatrixBooleanVariable(rows: int, cols: int, name: str) -> numpy.ndarray[object[m, n]]: ...
def MakeMatrixContinuousVariable(rows: int, cols: int, name: str) -> numpy.ndarray[object[m, n]]: ...
def MakeMatrixVariable(rows: int, cols: int, name: str, type: Variable.Type = ...) -> numpy.ndarray[object[m, n]]: ...
def MakeVectorBinaryVariable(rows: int, name: str) -> numpy.ndarray[object[m, 1]]: ...
def MakeVectorBooleanVariable(rows: int, name: str) -> numpy.ndarray[object[m, 1]]: ...
def MakeVectorContinuousVariable(rows: int, name: str) -> numpy.ndarray[object[m, 1]]: ...
def MakeVectorVariable(rows: int, name: str, type: Variable.Type = ...) -> numpy.ndarray[object[m, 1]]: ...
def MonomialBasis(vars: numpy.ndarray[object[m, 1]], degree: int) -> numpy.ndarray[object[m, 1]]: ...
def OddDegreeMonomialBasis(vars: Variables, degree: int) -> numpy.ndarray[object[m, 1]]: ...
@overload
def ReplaceBilinearTerms(e: Expression, x: numpy.ndarray[object[m, 1]], y: numpy.ndarray[object[m, 1]], W: numpy.ndarray[object[m, n], flags.f_contiguous]) -> Expression: ...
@overload
def ReplaceBilinearTerms(e, x, y, W) -> Any: ...
@overload
def Substitute(m: numpy.ndarray[object[m, n]], subst: dict[Variable, Expression]) -> numpy.ndarray[object[m, n]]: ...
@overload
def Substitute(m: numpy.ndarray[object[m, n]], var: Variable, e: Expression) -> numpy.ndarray[object[m, n]]: ...
@overload
def Substitute(e: Expression, subs: dict[Variable, SinCos]) -> Expression: ...
@overload
def Substitute(m: numpy.ndarray[object[m, n]], subs: dict[Variable, SinCos]) -> numpy.ndarray[object[m, n]]: ...
def SubstituteStereographicProjection(e: Polynomial, sin_cos: list[SinCos], t: numpy.ndarray[object[m, 1]]) -> RationalFunction: ...
def TaylorExpand(f: Expression, a: dict[Variable, float], order: int) -> Expression: ...
@overload
def ToLatex(e: Expression, precision: int = ...) -> str: ...
@overload
def ToLatex(f: Formula, precision: int = ...) -> str: ...
@overload
def ToLatex(M: numpy.ndarray[object[m, n]], precision: int = ...) -> str: ...
@overload
def ToLatex(M: numpy.ndarray[numpy.float64[m, n]], precision: int = ...) -> str: ...
def abs(arg0: Expression) -> Expression: ...
def acos(arg0: Expression) -> Expression: ...
def arccos(arg0: Expression) -> Expression: ...
def arcsin(arg0: Expression) -> Expression: ...
def arctan(arg0: Expression) -> Expression: ...
def arctan2(y: Expression, x: Expression) -> Expression: ...
def asin(arg0: Expression) -> Expression: ...
def atan(arg0: Expression) -> Expression: ...
def atan2(y: Expression, x: Expression) -> Expression: ...
def ceil(arg0: Expression) -> Expression: ...
def cos(arg0: Expression) -> Expression: ...
def cosh(arg0: Expression) -> Expression: ...
def exp(arg0: Expression) -> Expression: ...
def floor(arg0: Expression) -> Expression: ...
def forall(vars: Variables, f: Formula) -> Formula: ...
@overload
def if_then_else(f_cond: Formula, e_then: Expression, e_else: Expression) -> Expression: ...
@overload
def if_then_else(cond, expr_then, expr_else) -> Any: ...
def intersect(vars1: Variables, vars2: Variables) -> Variables: ...
def inv(arg0: numpy.ndarray[object[m, n]]) -> numpy.ndarray[object[m, n]]: ...
@overload
def isnan(x: Expression) -> Formula: ...
@overload
def isnan(e: Expression) -> Formula: ...
@overload
def isnan(e) -> Any: ...
@overload
def isnan(e) -> Any: ...
def log(arg0: Expression) -> Expression: ...
def logical_not(arg0: Formula) -> Formula: ...
def max(arg0: Expression, arg1: Expression) -> Expression: ...
def min(arg0: Expression, arg1: Expression) -> Expression: ...
@overload
def positive_semidefinite(m: numpy.ndarray[object[m, n], flags.f_contiguous]) -> Formula: ...
@overload
def positive_semidefinite(m) -> Any: ...
@overload
def pow(arg0: Expression, arg1: float) -> Expression: ...
@overload
def pow(arg0: Expression, arg1: Expression) -> Expression: ...
def sin(arg0: Expression) -> Expression: ...
def sinh(arg0: Expression) -> Expression: ...
def sqrt(arg0: Expression) -> Expression: ...
def tan(arg0: Expression) -> Expression: ...
def tanh(arg0: Expression) -> Expression: ...
def uninterpreted_function(name: str, arguments: list[Expression]) -> Expression: ...



"""
Provides general visualization utilities. This is NOT related to `rendering`.
"""

from tempfile import NamedTemporaryFile

from pydrake.common import temp_directory

# TODO(eric.cousineau): Move `plot_graphviz` to something more accessible to
# `call_python_client`.

# Use a global variable here because some calls to IPython will actually cause
# an interpreter to be created.  This file needs to be imported BEFORE that
# happens.
try:
    # IPython module may not be available; in which case, we're definitely
    # *not* running as a notebook.
    from IPython import get_ipython
    from IPython.display import SVG, display

    running_as_notebook = get_ipython() and hasattr(get_ipython(), "kernel")
except ModuleNotFoundError:
    running_as_notebook = False


def _plt():
    """Deferred import of plt, so pydrake.all importing doesn't bomb out."""
    # On Ubuntu the Debian package python3-tk is a recommended (but not
    # required) dependency of python3-matplotlib; help users understand
    # that by providing a nicer message upon a failure to import.
    result = None
    try:
        import matplotlib.pyplot as __plt

        result = __plt
    except ImportError as e:
        if e.name == "tkinter":
            result = None
        else:
            raise
    if result is None:
        raise NotImplementedError(
            "On Ubuntu when using the default pyplot configuration (i.e.,"
            " the TkAgg backend) you must 'sudo apt install python3-tk' to"
            " obtain Tk support. Alternatively, you may set MPLBACKEND to"
            " something else (e.g., Qt5Agg)."
        )
    return result


def _pydot():
    """Deferred import of pyplot, so pydrake.all importing doesn't bomb out."""
    import pydot as __pydot

    return __pydot


def plot_graphviz(dot_text):
    """Renders a DOT graph in matplotlib."""
    plt = _plt()
    pydot = _pydot()

    # @ref https://stackoverflow.com/a/18522941/7829525
    # Tried (reason ignored): pydotplus (`pydot` works), networkx
    # (`read_dot` does not work robustly?), pygraphviz (coupled with
    # `networkx`).
    g = pydot.graph_from_dot_data(dot_text)
    if isinstance(g, list):
        # Per Ioannis's follow-up comment in the above link, in pydot >= 1.2.3
        # `graph_from_dot_data` returns a list of graphs.
        # Handle this case for now.
        assert len(g) == 1
        g = g[0]
    if running_as_notebook:
        return display(SVG(g.create_svg()))
    else:
        f = NamedTemporaryFile(suffix=".png", dir=temp_directory())
        g.write_png(f.name)
        plt.axis("off")

        return plt.imshow(plt.imread(f.name), aspect="equal")


def plot_system_graphviz(system, **kwargs):
    """Renders a System's Graphviz representation in `matplotlib`."""
    return plot_graphviz(system.GetGraphvizString(**kwargs))



"""
Provides some general visualization tools using matplotlib.  This is not
related to `rendering`.
"""

import matplotlib as mpl
import numpy as np

from pydrake.solvers import MathematicalProgram, Solve
from pydrake.symbolic import Evaluate, Jacobian, Polynomial


def plot_sublevelset_quadratic(ax, A, b=[0, 0], c=0, vertices=51, **kwargs):
    """
    Plots the 2D ellipse representing x'Ax + b'x + c <= 1, e.g.
    the one sub-level set of a quadratic form.

    Args:
        ax:     the matplotlib axis to receive the plot
        A:      an 2x2 PSD numpy array.
        b:      broadcastable to a 2x1 array
        c:      scalar
        vertices: number of sample points along the boundary
        kwargs: are passed to the matplotlib fill method

    Returns:
        the return values from matplotlib's fill command.
    """
    # TODO(russt): Add project_quadratic_form and slice_quadratic_form
    #  methods to drake::math and recommend users call them first to plot
    #  higher-dimensional ellipsoids.
    assert isinstance(ax, mpl.axes.Axes)
    A = np.asarray(A)
    assert A.shape == (2, 2), "A must be 2x2"

    # Note: this is closely related to
    # drake::math::DecomposePositiveQuadraticForm, but does not require the
    # function to be positive for all x.

    # f = x'Ax + b'x + c
    # dfdx = x'(A+A') + b'
    # H = .5*(A+A')   # aka, the symmetric part of A (note: x'Hx = x'Ax)
    # dfdx = 0 => xmin = -inv(A+A')*b = -.5 inv(H)*b
    H = 0.5 * (A + A.T)
    xmin = np.linalg.solve(-2 * H, np.reshape(b, (2, 1)))
    fmin = -xmin.T.dot(H).dot(xmin) + c  # since b = -2*H*xmin
    assert fmin <= 1, (
        "The minimum value is > 1; there is no sub-level set to plot"
    )

    # To plot the contour at f = (x-xmin)'H(x-xmin) + fmin = 1,
    # we make a circle of values y, such that: y'y = 1-fmin,
    th = np.linspace(0, 2 * np.pi, vertices)
    Y = np.sqrt(1 - fmin) * np.vstack([np.sin(th), np.cos(th)])
    # then choose L'*(x - xmin) = y, where H = LL'.
    L = np.linalg.cholesky(H)
    X = np.tile(xmin, vertices) + np.linalg.inv(np.transpose(L)).dot(Y)

    return ax.fill(X[0, :], X[1, :], **kwargs)


def plot_sublevelset_expression(ax, e, vertices=51, **kwargs):
    """
    Plots the 2D sub-level set e(x) <= 1, which must contain the origin.

    Args:
        ax:       the matplotlib axis to receive the plot
        e:        a symbolic expression in two variables
        vertices: number of sample points along the boundary
        kwargs:   are passed to the matplotlib fill method

    Returns:
        the return values from matplotlib's fill command.
    """

    x = list(e.GetVariables())
    assert len(x) == 2, "e must be an expression in two variables"

    # Handle the special case where e is a degree 2 polynomial.
    if e.is_polynomial():
        p = Polynomial(e)
        if p.TotalDegree() == 2:
            env = {a: 0 for a in x}
            c = e.Evaluate(env)
            e1 = e.Jacobian(x)
            b = Evaluate(e1, env)
            e2 = Jacobian(e1, x)
            A = 0.5 * Evaluate(e2, env)
            return plot_sublevelset_quadratic(ax, A, b, c, vertices, **kwargs)

    # Find the level-set in polar coordinates, by sampling theta and
    # root-finding (on the scalar expression) to find a rplus and rminus.

    Xplus = np.empty((2, vertices))
    Xminus = np.empty((2, vertices))
    i = 0
    for theta in np.linspace(0, np.pi, vertices):
        prog = MathematicalProgram()
        r = prog.NewContinuousVariables(1, "r")[0]
        env = {x[0]: r * np.cos(theta), x[1]: r * np.sin(theta)}
        scalar = e.Substitute(env)
        b = prog.AddBoundingBoxConstraint(0, np.inf, r)
        prog.AddConstraint(scalar == 1)
        prog.AddQuadraticCost([1], [0], [r])
        prog.SetInitialGuess(r, 0.1)  # or anything non-zero.
        result = Solve(prog)
        assert result.is_success(), "Failed to find the level set"
        rplus = result.GetSolution(r)
        Xplus[0, i] = rplus * np.cos(theta)
        Xplus[1, i] = rplus * np.sin(theta)
        b.evaluator().UpdateLowerBound([-np.inf])
        b.evaluator().UpdateUpperBound([0])
        prog.SetInitialGuess(r, -0.1)  # or anything non-zero.
        result = Solve(prog)
        assert result.is_success(), "Failed to find the level set"
        rminus = result.GetSolution(r)
        Xminus[0, i] = rminus * np.cos(theta)
        Xminus[1, i] = rminus * np.sin(theta)
        i = i + 1

    return ax.fill(
        np.hstack((Xplus[0, :], Xminus[0, :])),
        np.hstack((Xplus[1, :], Xminus[1, :])),
        **kwargs,
    )



r"""The ``model_visualizer`` program displays a model file (e.g., ``*.sdf``) in
Drake's built-in visualizers (MeshCat and/or Meldis). When viewing in MeshCat,
joint sliders to posture the model are available by clicking on "Open Controls"
in the top right corner.

If the loaded model file is changed, it can be reloaded by pressing
the "Reload Model Files" button, which will attempt to maintain slider values
once reloading is finished. To exit, press the "Stop Running" button or press
the Escape key.

This command-line module is provided for convenience, but the feature is also
available via the library class ``pydrake.visualization.ModelVisualizer``.

From a Drake source build, run this module as::

  bazel run //tools:model_visualizer -- --help

From a Drake binary release (including pip releases), run this module as::

  python3 -m pydrake.visualization.model_visualizer --help

For binary releases (except for pip) there is also a shortcut available as::

  /opt/drake/bin/model_visualizer

Refer to the instructions printed by ``--help`` for additional details.

An example of viewing an iiwa model file::

  python3 -m pydrake.visualization.model_visualizer --open-window \
        package://drake_models/iiwa_description/sdf/iiwa7_with_box_collision.sdf

This program respects the ``ROS_PACKAGE_PATH``; if your model uses external
resources then you will need to set that environment variable.
"""

import argparse
import logging
import os
from pathlib import Path
import textwrap

from pydrake.common import configure_logging as _configure_logging
from pydrake.visualization._model_visualizer import (
    ModelVisualizer as _ModelVisualizer,
)


def _main():
    # Use a few color highlights for the user's terminal output.
    _configure_logging()
    logging.addLevelName(logging.INFO, "\033[36mINFO\033[0m")
    logging.addLevelName(logging.WARNING, "\033[33mWARNING\033[0m")
    logging.addLevelName(logging.ERROR, "\033[31mERROR\033[0m")

    # Prepare to parse arguments.
    args_parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # Many of our command line arguments map directly onto named arguments to
    # the ModelVisualizer constructor. We'll obey those constructor defaults
    # as we define the argparse defaults.
    defaults = _ModelVisualizer._get_constructor_defaults()

    args_parser.add_argument(
        "filename",
        nargs="+",
        type=str,
        help="Filesystem path to an SDFormat, URDF, OBJ, or DMD file; "
        "or a package:// URL to use a ROS package path.",
    )

    assert defaults["browser_new"] is False
    args_parser.add_argument(
        "-w",
        "--open-window",
        dest="browser_new",
        action="store_true",
        help="Open the MeshCat display in a new browser window.",
    )
    assert defaults["pyplot"] is False
    args_parser.add_argument(
        "--pyplot",
        action="store_true",
        help="Open a pyplot figure for rendering using "
        "PlanarSceneGraphVisualizer.",
    )
    # TODO(russt): Consider supporting the PlanarSceneGraphVisualizer
    #  options as additional arguments.
    assert defaults["visualize_frames"] is False
    args_parser.add_argument(
        "--visualize_frames",
        action="store_true",
        help="Visualize the frames as triads for all links.",
    )
    assert defaults["show_rgbd_sensor"] is False
    args_parser.add_argument(
        "--show_rgbd_sensor",
        action="store_true",
        help="Add and show an RgbdSensor. At the moment, the image display "
        "uses a native window so will not work in a remote or cloud "
        "runtime environment.",
    )
    assert defaults["environment_map"] == Path()
    args_parser.add_argument(
        "--environment_map",
        default=Path(),
        type=Path,
        help="Filesystem path to an image to be used as an environment map. "
        "It must be an image type normally used by your browser (e.g., "
        ".jpg, .png, etc.). HDR images are not supported yet.",
    )
    args_parser.add_argument(
        "--no_lights",
        action="store_true",
        help="If set, meshcat's lights will be disabled, similarly if "
        "--show_rgbd_sensor is enabled, the sensor will have no visible "
        "lights. This should be used in conjunction with --environment_map.",
    )
    args_parser.add_argument(
        "--compliance_type",
        default=defaults["compliance_type"],
        help=textwrap.dedent("""Overrides the DefaultProximityProperties
        setting with same name. Can be set to either 'rigid' or 'compliant' for
        hydroelastic contact, or 'undefined' to use point contact.  When a
        model file doesn't say something more specific for the hydroelastic
        compliance mode, this default will take effect.  In the common case of
        model files that have not been customized for Drake, this is a
        convenient way to visualize what collisions would look like under the
        given hydroelastic mode."""),
    )

    args_parser.add_argument(
        "--triad_length",
        type=float,
        dest="triad_length",
        default=defaults["triad_length"],
        help="Triad length for frame visualization.",
    )
    args_parser.add_argument(
        "--triad_radius",
        type=float,
        dest="triad_radius",
        default=defaults["triad_radius"],
        help="Triad radius for frame visualization.",
    )
    args_parser.add_argument(
        "--triad_opacity",
        type=float,
        dest="triad_opacity",
        default=defaults["triad_opacity"],
        help="Triad opacity for frame visualization.",
    )
    args_parser.add_argument(
        "-q",
        "--position",
        dest="position",
        type=float,
        nargs="+",
        default=[],
        help="A list of positions which must be the same length as the number "
        "of positions in the sdf models.  Note that most models have a "
        "floating-base joint by default (unless the sdf explicitly welds "
        "the base to the world, and so have 7 positions corresponding to "
        "the quaternion representation of that floating-base position).",
    )

    args_parser.add_argument(
        "--loop_once",
        action="store_true",
        help="Run the evaluation loop once and then quit.",
    )
    args = args_parser.parse_args()

    if "BUILD_WORKSPACE_DIRECTORY" in os.environ:
        os.chdir(os.environ["BUILD_WORKING_DIRECTORY"])

    visualizer = _ModelVisualizer(
        visualize_frames=args.visualize_frames,
        show_rgbd_sensor=args.show_rgbd_sensor,
        triad_length=args.triad_length,
        triad_radius=args.triad_radius,
        triad_opacity=args.triad_opacity,
        browser_new=args.browser_new,
        pyplot=args.pyplot,
        environment_map=args.environment_map,
        no_lights=args.no_lights,
        compliance_type=args.compliance_type,
    )
    package_map = visualizer.package_map()
    package_map.PopulateFromRosPackagePath()
    for item in args.filename:
        if item.startswith("package://"):
            visualizer.AddModels(url=item)
        else:
            visualizer.AddModels(filename=Path(item))

    # Not all model files are compatible with computing dynamics (e.g., they
    # might have zero-mass floating bodies). We'll try computing to see if it
    # works, but if not then we'll need to turn off contact visualization.
    try:
        visualizer.Finalize()
    except RuntimeError:
        logging.warning("Contact results cannot be visualized")
        visualizer._publish_contacts = False
        visualizer._reload()

    visualizer.Run(position=args.position, loop_once=args.loop_once)


if __name__ == "__main__":
    _main()



"""
Bindings for ``math``, including overloads for scalar types and basic SE(3)
representations.

Note that arrays of symbolic scalar types, such as ``Variable`` and
``Expression``, are exposed using ``ndarray[object]``, and as such logical
operations are constrained to return boolean values given NumPy's
implementation; this is not desirable, as one should really get a ``Formula``
object. As a workaround, this module provides the following vectorized
operators, following suit with the ``operator`` builtin module:
``lt``, ``le``, ``eq``, ``ne``, ``ge``, and ``gt``.

As an example::

    >>> x = np.array([Variable("x0"), Variable("x1")])
    >>> y = np.array([Variable("y0"), Variable("y1")])
    >>> x >= y
    # This should throw a RuntimeError
    >>> ge(x, y)
    array([<Formula "(x0 >= y0)">, <Formula "(x1 >= y1)">], dtype=object)

"""

# See `ExecuteExtraPythonCode` in `pydrake_pybind.h` for usage details and
# rationale.

# ruff: noqa: F821 (undefined-name). This file is only a fragment.

import functools
import operator

import numpy as np

from pydrake.autodiffutils import AutoDiffXd as _AutoDiffXd
from pydrake.common import (
    _MangledName,
)
from pydrake.common import (
    pretty_class_name as _pretty_class_name,
)
import pydrake.symbolic as _sym

_sym_cls_list = (
    _sym.Expression,
    _sym.Variable,
)


def _is_elementwise_comparison_error(e):
    return (
        # Newer error message (numpy >= 1.16.0).
        "elementwise comparison failed" in str(e)
        # Older error messages.
        or "elementwise == comparison failed" in str(e)
        or "elementwise != comparison failed" in str(e)
    )


def _best_effort_rich_compare(a, b, *, oper):
    try:
        return oper(a, b)
    except RuntimeError as e:
        if "not call `__bool__` / `__nonzero__` on `Formula`" in str(e):
            if isinstance(a, _sym_cls_list):
                return oper(a, _sym.Expression(b))
            elif isinstance(b, _sym_cls_list):
                return oper(_sym.Expression(a), b)
        raise
    except DeprecationWarning as e:
        # N.B. This is only appears to be triggered for symbolic types.
        if _is_elementwise_comparison_error(e):
            if isinstance(a, np.generic):
                a = float(a)
            elif isinstance(b, np.generic):
                b = float(b)
            else:
                raise RuntimeError("Unexpected condition")
            return oper(a, b)
        raise


def _drake_vectorize(oper, *, doc):
    wrapped = functools.partial(_best_effort_rich_compare, oper=oper)
    return np.vectorize(wrapped, doc=doc)


# As mentioned in top-level, add generic logical operators as ufuncs so that we
# may do comparisons on arrays of any scalar type, without restriction on the
# output type. These are added solely to work around #8315, where arrays of
# Expression can't use direct logical operators (e.g. `<=`) since the output
# type is not bool.
# N.B. Defined in order listed in Python documentation:
# https://docs.python.org/3.6/library/operator.html
lt = _drake_vectorize(operator.lt, doc="Drake's vectorized `lt`")
le = _drake_vectorize(operator.le, doc="Drake's vectorized `le`")
eq = _drake_vectorize(operator.eq, doc="Drake's vectorized `eq`")
ne = _drake_vectorize(operator.ne, doc="Drake's vectorized `ne`")
ge = _drake_vectorize(operator.ge, doc="Drake's vectorized `ge`")
gt = _drake_vectorize(operator.gt, doc="Drake's vectorized `gt`")

# The following values are defined for testing.
_OPERATORS = (lt, le, eq, ne, ge, gt)
# - Equivalent expression when operands are reversed.
_OPERATORS_REVERSE = {
    lt: gt,
    le: ge,
    eq: eq,
    ne: ne,
    ge: le,
    gt: lt,
}


def _indented_repr(o):
    """Returns repr(o), with any lines beyond the first one indented +2."""
    return repr(o).replace("\n", "\n  ")


def _roll_pitch_yaw_repr(rpy):
    return (
        f"{_pretty_class_name(type(rpy))}("
        f"roll={repr(rpy.roll_angle())}, "
        f"pitch={repr(rpy.pitch_angle())}, "
        f"yaw={repr(rpy.yaw_angle())})"
    )


def _rotation_matrix_repr(R):
    M = R.matrix().tolist()
    return (
        f"{_pretty_class_name(type(R))}([\n"
        f"  {_indented_repr(M[0])},\n"
        f"  {_indented_repr(M[1])},\n"
        f"  {_indented_repr(M[2])},\n"
        f"])"
    )


def _rigid_transform_repr(X):
    return (
        f"{_pretty_class_name(type(X))}(\n"
        f"  R={_indented_repr(X.rotation())},\n"
        f"  p={_indented_repr(X.translation().tolist())},\n"
        f")"
    )


def _add_repr_functions():
    for T in [float, _AutoDiffXd, _sym.Expression]:
        RollPitchYaw_[T].__repr__ = _roll_pitch_yaw_repr
        RotationMatrix_[T].__repr__ = _rotation_matrix_repr
        RigidTransform_[T].__repr__ = _rigid_transform_repr


_add_repr_functions()


def __getattr__(name):
    """Rewrites requests for Foo[bar] into their mangled form, for backwards
    compatibility with unpickling.
    """
    return _MangledName.module_getattr(
        module_name=__name__, module_globals=globals(), name=name
    )



"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_point(object):
    """ A representation of a 3D point. """

    __slots__ = ["x", "y", "z"]

    __typenames__ = ["double", "double", "double"]

    __dimensions__ = [None, None, None]

    def __init__(self):
        self.x = 0.0
        """ LCM Type: double """
        self.y = 0.0
        """ LCM Type: double """
        self.z = 0.0
        """ LCM Type: double """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_point._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">ddd", self.x, self.y, self.z))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_point._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_point._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_point()
        self.x, self.y, self.z = struct.unpack(">ddd", buf.read(24))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_point in parents: return 0
        tmphash = (0x573f2fdd2f76508f) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_point._packed_fingerprint is None:
            lcmt_point._packed_fingerprint = struct.pack(">Q", lcmt_point._get_hash_recursive([]))
        return lcmt_point._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_point._get_packed_fingerprint())[0]




"""
Python bindings for
`Drake: Model-Based Design and Verification for Robotics
<https://drake.mit.edu/>`_.

This Python API documentation is a work in progress. Most of it is generated
automatically from the C++ API documentation, and thus may have some
C++-specific artifacts. For general overview and API documentation, please see
the `Doxygen C++ Documentation
<https://drake.mit.edu/doxygen_cxx/index.html>`_.

For examples and tutorials that tie in and use this API, please see
`here <https://drake.mit.edu/#tutorials-and-examples>`_.
"""

import functools
import os
import sys
import warnings


def getDrakePath():
    # Compatibility alias.
    return os.path.abspath(common.GetDrakePath())


def _execute_extra_python_code(m, use_subdir: bool = False):
    # See `ExecuteExtraPythonCode` in `pydrake_pybind.h` for usage details and
    # rationale.
    if m.__name__ not in sys.modules:
        # N.B. This is necessary for C++ extensions in Python 3.
        sys.modules[m.__name__] = m
    module_path = m.__name__.split(".")
    if len(module_path) == 1:
        raise RuntimeError(
            "ExecuteExtraPythonCode cannot be used with the top-level "
            f"module `{m.__name__}`. If you are writing modules in a "
            "downstream project, please review this thread and ensure your "
            "import is correct: https://stackoverflow.com/a/57858822/7829525"
        )
    top_module_name = module_path[0]
    top_module_dir = os.path.dirname(sys.modules[top_module_name].__file__)
    if use_subdir:
        mid_module_names = module_path[1:]
    else:
        mid_module_names = module_path[1:-1]
    base_module_name = module_path[-1]
    if base_module_name.startswith("_"):
        # Do not repeat leading `_`.
        extra_module_name = f"{base_module_name}_extra.py"
    else:
        extra_module_name = f"_{base_module_name}_extra.py"
    extra_path = [top_module_dir] + mid_module_names + [extra_module_name]
    extra_filename = os.path.join(*extra_path)
    with open(extra_filename) as f:
        _code = compile(f.read(), extra_filename, "exec")
        exec(_code, m.__dict__, m.__dict__)


def _setattr_kwargs(obj, kwargs):
    # For `ParamInit` in `pydrake_pybind.h`.
    for name, value in kwargs.items():
        setattr(obj, name, value)


def _import_cc_module_vars(
    cc_module,
    py_module_name,
    *,
    include_private=False,
):
    # Imports the cc_module's public symbols into the named py module
    # (py_module_name), resetting their __module__ to be the py module in the
    # process.
    # Returns a list[str] of the public symbol names imported.
    py_module = sys.modules[py_module_name]
    var_list = []
    for name, value in cc_module.__dict__.items():
        if name.startswith("_") and not include_private:
            continue
        if name.startswith("__"):
            continue
        if getattr(value, "__module__", None) == cc_module.__name__:
            value.__module__ = py_module_name
        setattr(py_module, name, value)
        var_list.append(name)
    return var_list


@functools.lru_cache
def _is_building_documentation():
    """Returns True iff pydrake is being imported by the website documentation
    build process (i.e., Sphinx). We use this to adjust our code to be more
    documentation-suitable if so.
    """
    return "DRAKE_IS_BUILDING_DOCUMENTATION" in os.environ


class _DrakeImportWarning(Warning):
    pass


_RTLD_GLOBAL_WARNING = r"""
You may have already (directly or indirectly) imported `torch` which uses
`RTLD_GLOBAL`. Using `RTLD_GLOBAL` may cause symbol collisions which manifest
themselves in bugs like "free(): invalid pointer". Please consider importing
`pydrake` (and related C++-wrapped libraries like `cv2`, `open3d`, etc.)
*before* importing `torch`. For more details, see:
https://github.com/pytorch/pytorch/issues/3059#issuecomment-534676459
"""


def _check_for_rtld_global_usages():
    # Naively check if `torch` is using RTLD_GLOBAL. For more information, see
    # the above _RTLD_GLOBAL_WARNING message, #12073, and #13707.
    torch = sys.modules.get("torch")
    if torch is None:
        return False
    # This symbol was introduced in v1.5.0 (pytorch@ddff4efa2).
    # N.B. Per investigation in #13707, it seems like torch==1.4.0 also plays
    # better with pydrake. However, we will keep our warning conservative.
    using_rtld_global = getattr(torch, "USE_RTLD_GLOBAL_WITH_LIBTORCH", True)
    if not using_rtld_global:
        return False
    init_file = getattr(torch, "__file__")
    if init_file.endswith(".pyc"):
        init_file = init_file[:-1]
    if not init_file.endswith(".py"):
        return False
    with open(init_file) as f:
        init_source = f.read()
    return "sys.setdlopenflags(_dl_flags.RTLD_GLOBAL" in init_source


if _check_for_rtld_global_usages():
    warnings.warn(
        _RTLD_GLOBAL_WARNING, category=_DrakeImportWarning, stacklevel=3
    )


# We specifically load `common` prior to loading any other pydrake modules,
# in order to a) get assertion configuration done as early as possible, and b)
# detect whether we are able to load the shared libraries. But importantly,
# we need to do this *after* defining the pure Python helper functions, above,
# since they are called while loading `pydrake.common` and things it loads.
try:
    from . import common
except ImportError as e:
    if "cannot open shared object file" in (e.msg or "") and "/pydrake/" in (
        e.path or ""
    ):
        message = """
Drake failed to load a required library. This could indicate an installation
problem, or that your system is missing required distro-provided packages.
Please refer to the installation instructions to ensure that all required
dependencies are installed.
"""
        # For wheel builds, we have a file with additional advice.
        wheel_doc = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "INSTALLATION"
        )
        if os.path.exists(wheel_doc):
            with open(wheel_doc) as f:
                message += f.read()
        message += """
For more information, please see https://drake.mit.edu/installation.html
"""
        print(message)
    raise

__all__ = ["common", "getDrakePath"]



"""Downloads, verifies, and extracts a remote archive file. Do not run this
program by hand; it is used by drake/multibody/parsing/package_map.cc to fetch
remote packages on demand into a per-user local cache (e.g., ~/.cache/...).

For example, with an output_dir of `~/.cache/drake/package_map/foo` this
program will create this directory and an adjacent file:

 ~/.cache/drake/package_map/foo/
 ~/.cache/drake/package_map/foo.README

It is not an error for the output_dir to already exist when this program
begins, nor for it to come into being at any point while this program is
running. In other words, it is safe for two downloaders to be running at the
same time with the same arguments.
"""

# N.B. The ONLY packages we're allowed to use here are the Python standard
# library. This program is run using the host OS's built-in Python interpreter,
# which might not offer anything more.
import hashlib
import json
import logging
import os
from pathlib import Path
import shutil
import sys
import tempfile
import urllib.parse
import urllib.request as request


def _fail(message):
    raise SystemExit(message)


def _run(
    *,
    temp_dir: Path,
    package_name: str,
    urls: list[str],
    sha256: str,
    output_dir: Path,
    archive_type: str = None,
    strip_prefix: str = None,
):
    """Runs the download and extract logic, assuming already-validated args.

    Args:
        temp_dir: A path to a private (and empty) scratch directory. The caller
            is responsible for cleaning it up; we don't need to.
        package_name: The ROS package name for this package.
        urls: A list of download URLs for this package. Each one will be tried
            in turn until one succeeds.
        sha256: The cryptographic checksum of the file to be downloaded, as a
            64-character hexadecimal string.
        output_dir: The root directory to extract into. Its parent must exist,
            but the output_dir itself generally will not already exist. This
            script will create the output_dir if and only if the download and
            extraction was successful, i.e., as a single atomic transaction.
        archive_type: The archive type of the downloaded file. When not given,
            the archive type is determined from the file extension of the URL.
        strip_prefix: A directory prefix to remove from the extracted files. In
            many cases, an archive will prefix all filenames with something
            like "package-v1.2.3/" so that it extracts into a convenient
            directory. This option will discard that common prefix when
            extracting the archive. It is an error if the archive does not
            contain any diectory with this prefix, but if there are files
            outside of this directory they will be silently discarded.
    """
    # Add a README nearby to help explain the fetched data.
    # To avoid stomping on concurrently-running download of the same package
    # from another process, we need to write into a tempfile and then rename
    # it into place as a final step. (Renames are atomic; writes are not.)
    readme = output_dir.with_name(output_dir.name + ".README")
    fd, temp_readme = tempfile.mkstemp(
        dir=readme.parent, prefix=f"temp.{readme.name}."
    )
    temp_readme = Path(temp_readme)
    with os.fdopen(fd, "w") as f:
        f.write(
            f"Our sibling directory ./{output_dir.name} contains "
            f"package://{package_name}.\n\n"
            f"It was downloaded by Drake from one of these URLs:\n"
        )
        for url in urls:
            f.write(f" {url}\n")
    # Fix the README's access permissions (to 'rw-r--r--' from 'rw-------').
    temp_readme.chmod(0o644)

    # Try each url in turn.
    success = False
    errors = []
    for url in urls:
        basename = urllib.parse.urlparse(url).path.split("/")[-1] or "empty"
        temp_filename = temp_dir / basename
        hasher = hashlib.sha256()
        with open(temp_filename, "wb") as f:
            try:
                with request.urlopen(url=url, timeout=30) as response:
                    while True:
                        data = response.read(4096)
                        if not data:
                            break
                        hasher.update(data)
                        f.write(data)
            except OSError as e:
                errors.append(f"Candidate {url} failed:\n{e}")
                continue
        download_sha256 = hasher.hexdigest()
        if download_sha256 == sha256:
            success = True
            break
        errors.append(
            f"Candidate {url} failed:\n"
            f"Checksum mismatch; was {download_sha256} but wanted {sha256}."
        )

    # Report in case no downloads succeeded.
    if not success:
        messages = "\n\n".join(errors)
        _fail(f"All downloads failed:\n\n{messages}")

    # Unpack and check that the strip_prefix was valid.
    unpack_dir = temp_dir / "unpack"
    shutil.unpack_archive(
        filename=temp_filename, extract_dir=unpack_dir, format=archive_type
    )
    unpack_package_dir = unpack_dir / (strip_prefix or "")
    if not unpack_package_dir.is_dir():
        _fail(
            f"The strip_prefix='{strip_prefix}' does not exist "
            f"within {basename}"
        )

    # Now comes the really tricky part. Moving a directory is atomic but only
    # on the same filesystem! We need to move it from the temporary directory
    # into the output directory (under a temporary name), and then atomically
    # rename it back to the real name. (In general, it's not possible to check
    # whether two existing paths are on the same filesystem, i.e., we can't
    # know whether unpack_dir can atomically rename to output_dir, so we'll
    # always take the long way around.)
    with tempfile.TemporaryDirectory(
        dir=output_dir.parent, prefix=f"temp.{output_dir.name}."
    ) as output_dir_temp:
        move_destination = Path(output_dir_temp) / "incoming"
        shutil.move(src=unpack_package_dir, dst=move_destination)
        try:
            move_destination.rename(output_dir)
        except OSError:
            # This can happen if some *other* downloader process succeeded with
            # its download & extract & move & rename while we were still only
            # partway done. If that occurs, then the other downloader succeeded
            # and we can declare "mission accomplished" for ourself, too.
            pass

    # Now that we've confirmed all of the download details were satisfactory,
    # we can move the README into place.
    temp_readme.rename(readme)

    # Double-check that we've met our victory condition: the output_dir that we
    # were supposed to create using the downloaded archive actually exists now.
    assert output_dir.is_dir()


def _wrapped_main(*, config_json):
    # Read our config file.
    with open(config_json, "r") as f:
        kwargs = json.load(f)

    # Sanity check the required arguments. We use assert (not exceptions)
    # because the package_map.cc code should guarantee these always pass.
    package_name = kwargs["package_name"]
    urls = kwargs["urls"]
    sha256 = kwargs["sha256"]
    output_dir = kwargs["output_dir"]
    assert len(package_name) > 0
    assert len(urls) > 0
    assert len(sha256) == 64
    int(f"0x{sha256}", 16)
    assert "/drake/package_map/" in output_dir, output_dir
    kwargs["output_dir"] = Path(output_dir).resolve()

    # Download and extract.
    with tempfile.TemporaryDirectory(prefix="drake_downloader_") as temp_dir:
        _run(temp_dir=Path(temp_dir), **kwargs)


def _main(argv):
    # We expect exactly two command-line arguments to this program:
    # - The input filename containing JSON data with our *actual* arguments.
    # - The output filename we should use to report error message text.
    json_filename, error_filename = argv
    with open(error_filename, "w", encoding="utf-8") as error_file:
        try:
            _wrapped_main(config_json=json_filename)
            returncode = 0
        except SystemExit as e:
            error_file.write(str(e))
            returncode = 1
        except Exception as e:
            error_file.write(str(e))
            returncode = 1
    sys.exit(returncode)


if __name__ == "__main__":
    logging.basicConfig()
    _main(sys.argv[1:])



"""Provides an example translation of `cart_pole_passive_simluation.cc`."""

import argparse

from pydrake.common import configure_logging
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph
from pydrake.systems.analysis import Simulator
from pydrake.systems.framework import DiagramBuilder
from pydrake.visualization import AddDefaultVisualization


def main():
    configure_logging()
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--target_realtime_rate",
        type=float,
        default=1.0,
        help="Desired rate relative to real time.  See documentation for "
        "Simulator::set_target_realtime_rate() for details.",
    )
    parser.add_argument(
        "--simulation_time",
        type=float,
        default=10.0,
        help="Desired duration of the simulation in seconds.",
    )
    parser.add_argument(
        "--time_step",
        type=float,
        default=0.0,
        help="If greater than zero, the plant is modeled as a system with "
        "discrete updates and period equal to this time_step. "
        "If 0, the plant is modeled as a continuous system.",
    )
    args = parser.parse_args()

    builder = DiagramBuilder()
    cart_pole, scene_graph = AddMultibodyPlantSceneGraph(
        builder=builder, time_step=args.time_step
    )
    Parser(builder=builder).AddModelsFromUrl(
        url="package://drake/examples/multibody/cart_pole/cart_pole.sdf"
    )
    cart_pole.Finalize()

    AddDefaultVisualization(builder=builder)
    diagram = builder.Build()

    diagram_context = diagram.CreateDefaultContext()
    cart_pole_context = cart_pole.GetMyMutableContextFromRoot(diagram_context)

    cart_pole.get_actuation_input_port().FixValue(cart_pole_context, 0)

    cart_slider = cart_pole.GetJointByName("CartSlider")
    pole_pin = cart_pole.GetJointByName("PolePin")
    cart_slider.set_translation(context=cart_pole_context, translation=0.0)
    pole_pin.set_angle(context=cart_pole_context, angle=2.0)

    simulator = Simulator(diagram, diagram_context)
    simulator.set_target_realtime_rate(args.target_realtime_rate)
    simulator.Initialize()
    simulator.AdvanceTo(args.simulation_time)


if __name__ == "__main__":
    main()



"""
Drake Gym
=========

THIS FEATURE IS EXPERIMENTAL.
As per our [guidelines](https://drake.mit.edu/stable.html)
for experimental code, development is ongoing and no guarantees against
deprecation are provided for any file under this directory.

Drake Gym is an implementation of Farama's "Gymnasium" interface for
reinforcement learning which uses a Drake `Simulator` as a
backend.  The Gym interface provided
by [the python `gym` module](https://pypi.org/project/gymnasium/)
simply models a time-stepped process with an action space,
a reward function, and some form of state observation.

Note that pydrake.gym is an optional component of pydrake, and will only work
when the gymnasium package is also installed. As such, the DrakeGymEnv and
related code is not available for import as part of `pydrake.all`.

A note on dependencies
----------------------

In order to use a Gym, code must implement a `gymnasium.Env` (in this case the
actual `Simulator` wrapped via `DrakeGymEnv`) and must run the Gym within an
RL engine of some sort (that's usually an algorithm chosen from
[Stable Baselines 3]
(https://stable-baselines3.readthedocs.io/en/master/index.html)
but could also be from `nevergrad` or some other source of gradient-free
optimizers).

Stable Baselines3 iteself is too large and too heavy of a dependency tree for
Drake to require itself; as such you will need to provide it yourself to use
these examples.  If you are going to train drake gym examples on your machine,
you should install Stable Baselines 3 (for instance,
`pip install stable_baselines3` inside of a virtual environment).  The
training examples will not run without it, and drake does not come with it
(Drake does include a subset of `stable_baselines3` for testing purposes, but
not enough to perform training).
"""

from ._drake_gym_env import DrakeGymEnv  # noqa: F401 (unused-import)

__all__ = [x for x in globals() if not x.startswith("_")]



import copy
import logging
import os
from pathlib import Path
import re
import sys
import time
from webbrowser import open as _webbrowser_open

import numpy as np

from pydrake.geometry import (
    Box,
    EnvironmentMap,
    EquirectangularMap,
    LightParameter,
    MeshcatCone,
    RenderEngineVtkParams,
    Rgba,
    StartMeshcat,
)
from pydrake.math import RigidTransform, RotationMatrix
from pydrake.multibody.meshcat import JointSliders
from pydrake.multibody.tree import (
    FixedOffsetFrame,
    FrameIndex,
    default_model_instance,
)
from pydrake.planning import RobotDiagramBuilder
from pydrake.systems.analysis import Simulator
from pydrake.systems.planar_scenegraph_visualizer import (
    ConnectPlanarSceneGraphVisualizer,
)
from pydrake.systems.sensors import (
    ApplyCameraConfig,
    CameraConfig,
)
from pydrake.visualization import (
    ApplyVisualizationConfig,
    VisualizationConfig,
)
from pydrake.visualization._triad import (
    AddFrameTriadIllustration,
)


class ModelVisualizer:
    """
    Visualizes models from a file or string buffer in MeshCat or Meldis.

    To use this class to visualize model(s), create an instance with
    any desired options, add any models, and then call Run()::

      visualizer = ModelVisualizer(browser_new=True)
      visualizer.AddModels(filename)
      visualizer.Run()

    The class also provides a `parser()` method to allow more complex
    Parser handling, e.g. adding a model from a string::

      visualizer.parser().AddModelsFromString(buffer_containing_model, 'sdf')

    This class may also be run as a standalone command-line tool using the
    ``pydrake.visualization.model_visualizer`` script, or via
    ``bazel run //tools:model_visualizer``.
    """

    # Note: this class uses C++ method names to ease future porting.

    def __init__(
        self,
        *,
        visualize_frames=False,
        triad_length=0.3,
        triad_radius=0.005,
        triad_opacity=0.9,
        publish_contacts=True,
        show_rgbd_sensor=False,
        browser_new=False,
        pyplot=False,
        meshcat=None,
        environment_map: Path = Path(),
        no_lights: bool = False,
        compliance_type: str = "undefined",
    ):
        """Initializes a ModelVisualizer.

        Args:
          visualize_frames: a flag that visualizes frames as triads for all
            links.
          triad_length: the length of visualization triads.
          triad_radius: the radius of visualization triads.
          triad_opacity: the opacity of visualization triads.
          publish_contacts: a flag for VisualizationConfig.
          show_rgbd_sensor: when True, adds an RgbdSensor to the scene and pops
             up a local preview window of the rgb image. At the moment, the
             image display uses a native window so will not work in a remote or
             cloud runtime environment.
          environment_map: Meshcat environment map filename.
          no_lights: optionally disable the lights in the render engine and
             meshcat. This is useful when using an environment map to assess
             the effect of the map.
          compliance_type: Overrides the DefaultProximityProperties setting
             with same name. Can be set to either "rigid" or "compliant" for
             hydroelastic contact, or "undefined" to use point contact.
             When a model file doesn't say something more specific for the
             hydroelastic compliance mode, this default will take effect.
             In the common case of model files that have not been customized
             for Drake, this is a convenient way to visualize what collisions
             would look like under the given hydroelastic mode.
          browser_new: a flag that will open the MeshCat display in a new
            browser window during Run().
          pyplot: a flag that will open a pyplot figure for rendering using
            PlanarSceneGraphVisualizer.

          meshcat: an existing Meshcat instance to re-use instead of creating
            a new instance. Useful in, e.g., Python notebooks.

        """
        self._visualize_frames = visualize_frames
        self._triad_length = triad_length
        self._triad_radius = triad_radius
        self._triad_opacity = triad_opacity
        self._publish_contacts = publish_contacts
        self._show_rgbd_sensor = show_rgbd_sensor
        self._browser_new = browser_new
        self._pyplot = pyplot
        self._meshcat = meshcat
        self._environment_map = environment_map
        self._no_lights = no_lights
        self._compliance_type = compliance_type

        # This is the list of loaded models, to enable the Reload button.
        # If set to None, it means that we won't support reloading because
        # the user might have added models outside of our purview. Each item
        # in the list contains whatever kwargs we passed to AddModels().
        self._added_models = list()

        # This is set to a non-None value iff our Meshcat has a reload button.
        self._reload_button_name = None

        # The builder is set to None during Finalize(), though during a Reload
        # it will be temporarily resurrected.
        self._builder = RobotDiagramBuilder()
        self._builder.parser().SetAutoRenaming(True)

        # Adjust the SceneGraph's compliance_type.
        old_config = self._builder.scene_graph().get_config()
        new_config = copy.deepcopy(old_config)
        new_config.default_proximity_properties.compliance_type = (
            self._compliance_type
        )
        self._builder.scene_graph().set_config(new_config)

        # The following fields are set non-None during Finalize().
        self._original_package_map = None
        self._diagram = None  # This will be a planning.RobotDiagram.
        self._sliders = None
        self._context = None

        # State necessary for self._render_if_necessary().
        self._last_camera_time = time.time()

    def _check_rep(self, *, finalized):
        """
        Checks that our self members are consistent with the provided expected
        state of finalization.
        """
        if not finalized:
            # The builder is alive.
            assert self._builder is not None
            # The meshcat might or might not exist yet.
            # Everything else is dead.
            assert self._original_package_map is None
            assert self._diagram is None
            assert self._sliders is None
            assert self._context is None
        else:
            # The builder is dead.
            assert self._builder is None
            # Everything else is alive.
            assert self._meshcat is not None
            assert self._original_package_map is not None
            assert self._diagram is not None
            assert self._sliders is not None
            assert self._context is not None

    @staticmethod
    def _get_constructor_defaults():
        """
        Returns a dict of the default values used in our constructor's named
        keyword arguments (for any non-None values); this helps our companion
        main() function share those same defaults.
        """
        result = dict()
        prototype = ModelVisualizer()
        for name in [
            "visualize_frames",
            "triad_length",
            "triad_radius",
            "triad_opacity",
            "publish_contacts",
            "show_rgbd_sensor",
            "browser_new",
            "pyplot",
            "environment_map",
            "compliance_type",
        ]:
            value = getattr(prototype, f"_{name}")
            assert value is not None
            result[name] = value
        return result

    def package_map(self):
        """
        Returns the PackageMap being used during parsing. Users should add
        entries here to be able to load models from non-Drake packages.

        This method cannot be used after Finalize is called.
        """
        if self._builder is None:
            raise ValueError("Finalize has already been called.")
        self._check_rep(finalized=False)
        # It's safe to let the user change the package map. We'll make a copy
        # of it during Finalize().
        return self._builder.parser().package_map()

    def parser(self):
        """
        (Advanced) Returns a Parser that will load models into this visualizer.

        Prefer to use package_map() and AddModels() to load models, instead of
        this method.

        Calling this method will disable the "Reload" button in the visualizer,
        because we can no longer determine the scope of what to reload.

        This method cannot be used after Finalize is called.
        """
        if self._builder is None:
            raise ValueError("Finalize has already been called.")
        self._check_rep(finalized=False)
        # We can't easily know what the user is going to do with the parser,
        # so we need to disable model reloading once they access it.
        self._added_models = None
        return self._builder.parser()

    def meshcat(self):
        """
        Returns the Meshcat object this visualizer is plugged into.
        If none was provided in the constructor, this creates one on demand.
        """
        if self._meshcat is None:
            self._meshcat = StartMeshcat()
        return self._meshcat

    def AddModels(self, filename: Path = None, *, url: str = None):
        """
        Adds all models found in an input file (or url).

        This can be called multiple times, until the object is finalized.

        Args:
          filename: the name of a file containing one or more models.
          url: the package:// URL containing one or more models.

        Exactly one of filename or url must be non-None.
        """
        if self._builder is None:
            raise ValueError("Finalize has already been called.")
        if sum([filename is not None, url is not None]) != 1:
            raise ValueError("Must provide either filename= or url=")
        if isinstance(filename, str):
            # Silently clean up the common mistake of passing the wrong type.
            filename = Path(filename)
        self._check_rep(finalized=False)
        if filename is not None:
            if filename.suffix == ".gltf":
                # TODO(jwnimmer-tri) Add glTF support to the detail_mesh_parser
                # (which involves computing its volume / inertia), instead of
                # doing the halfway flavor (visual-only) here.
                kwargs = self._wrap_gltf_as_visual(filename=filename)
            else:
                kwargs = dict(file_name=str(filename))
        else:
            assert url is not None
            if url.endswith(".gltf"):
                # TODO(jwnimmer-tri) Add glTF support to the detail_mesh_parser
                # (which involves computing its volume / inertia), instead of
                # doing the halfway flavor (visual-only) here.
                kwargs = self._wrap_gltf_as_visual(url=url)
            else:
                kwargs = dict(url=url)
        self._builder.parser().AddModels(**kwargs)
        if self._added_models is not None:
            self._added_models.append(kwargs)

    def _wrap_gltf_as_visual(self, *, filename: Path = None, url: str = None):
        """Given a filename xor url that refers to a glTF mesh, returns a dict
        of kwargs to Parser.AddModels which will load it as visual-only (i.e.,
        without collision geometry)"""
        assert sum([filename is not None, url is not None]) == 1
        if filename is not None:
            package_name = re.sub(r"[^A-Za-z0-9_]", "", str(filename.parent))
            self._builder.parser().package_map().Add(
                package_name=package_name,
                package_path=str(filename.parent),
            )
            url = f"package://{package_name}/{filename.name}"
        assert url is not None
        name = Path(url.split("/")[-1]).stem
        file_type = "sdf"
        file_contents = f"""
        <?xml version="1.0"?>
        <sdf version="1.12">
          <model name="{name}">
            <link name="body">
              <visual name="visual">
                <geometry>
                  <mesh>
                    <uri>{url}</uri>
                  </mesh>
                </geometry>
              </visual>
            </link>
          </model>
        </sdf>
        """
        return dict(
            file_type=file_type,
            file_contents=file_contents,
        )

    def Finalize(self, position=None):
        """
        Finalizes the object and sets up the visualization using the provided
        options once models have been added.

        Args:
          position: a list of slider positions for the model(s); must match
            the number of positions in all model(s), including the 7 positions
            corresponding to any model bases that aren't welded to the world.
        """
        self._check_rep(finalized=False)
        if self._builder is None:
            raise RuntimeError("Finalize has already been called.")

        if self._visualize_frames:
            # Find all the frames (except the world frame) and draw them.
            # The frames are drawn using the configured length.
            for i in range(1, self._builder.plant().num_frames()):
                AddFrameTriadIllustration(
                    plant=self._builder.plant(),
                    scene_graph=self._builder.scene_graph(),
                    frame_index=FrameIndex(i),
                    length=self._triad_length,
                    radius=self._triad_radius,
                    opacity=self._triad_opacity,
                )

        # Add a model to provide a pose-able anchor for the camera.
        if self._show_rgbd_sensor:
            sensor_offset_frame = self._builder.plant().AddFrame(
                FixedOffsetFrame(
                    name="$rgbd_sensor_offset",
                    P=self._builder.plant().world_frame(),
                    X_PF=RigidTransform.Identity(),
                    model_instance=default_model_instance(),
                )
            )
            sensor_body = self._builder.plant().AddRigidBody(
                name="$rgbd_sensor_body",
                model_instance=default_model_instance(),
            )
            self._builder.plant().WeldFrames(
                frame_on_parent_F=sensor_offset_frame,
                frame_on_child_M=sensor_body.body_frame(),
            )

        # We're not going to step time, so we don't want output port sampling.
        self._builder.plant().SetUseSampledOutputPorts(False)

        self._builder.plant().Finalize()

        # (Re-)initialize the meshcat instance, creating one if needed.
        self.meshcat()
        self._meshcat.Delete()
        self._meshcat.DeleteAddedControls()

        if self._no_lights:
            self._meshcat.SetProperty("/Lights", "visible", False)

        if self._environment_map.is_file():
            self._meshcat.SetEnvironmentMap(self._environment_map)

        # We want to place the Reload Model Files button far away from the
        # Stop Running button, hence the work to do this here.
        if self._added_models:
            self._reload_button_name = "Reload Model Files"
            self._meshcat.AddButton(self._reload_button_name)

        # Connect to meldis and meshcat.
        # Meldis and meshcat provide simultaneous visualization of
        # illustration and proximity geometry.
        ApplyVisualizationConfig(
            config=VisualizationConfig(
                publish_contacts=self._publish_contacts,
                enable_alpha_sliders=True,
            ),
            plant=self._builder.plant(),
            scene_graph=self._builder.scene_graph(),
            builder=self._builder.builder(),
            meshcat=self._meshcat,
        )

        # Add a render camera so we can show role=perception images. The
        # sensor is affixed to the world frame and we'll modify that pose
        # below.
        if self._show_rgbd_sensor:
            camera_config = CameraConfig(width=1440, height=1080)
            camera_config.name = "preview"
            camera_config.X_PB.base_frame = "$rgbd_sensor_body"
            camera_config.z_far = 3  # Show 3m of frustum.
            camera_config.fps = 1.0  # Ignored -- we're not simulating.
            # The meshcat default field of view is 75 degrees. We want the two
            # images to match.
            camera_config.focal = CameraConfig.FovDegrees(y=75)
            is_unit_test = "TEST_SRCDIR" in os.environ
            if not is_unit_test:
                # Pop up a local window.
                camera_config.show_rgb = True
                camera_config.renderer_class = RenderEngineVtkParams()
                camera_config.renderer_class.exposure = 1
                if self._environment_map.is_file():
                    camera_config.renderer_class.environment_map = (
                        EnvironmentMap(
                            skybox=True,
                            texture=EquirectangularMap(
                                path=str(self._environment_map)
                            ),
                        )
                    )
                if self._no_lights:
                    # We can only disable *all* lights by creating a light with
                    # zero intensity.
                    camera_config.renderer_class.lights = [
                        LightParameter(intensity=0)
                    ]
                if "darwin" not in sys.platform:
                    camera_config.renderer_class.backend = "GLX"
            ApplyCameraConfig(
                config=camera_config, builder=self._builder.builder()
            )
            camera_sensor = self._builder.builder().GetSubsystemByName(
                "rgbd_sensor_preview"
            )
            camera_publisher = self._builder.builder().GetSubsystemByName(
                "LcmPublisherSystem(DRAKE_RGBD_CAMERA_IMAGES_preview)"
            )
            # Export the preview camera image output port for later use.
            self._builder.builder().ExportOutput(
                camera_sensor.GetOutputPort("color_image"), "preview_image"
            )
            # Disable LCM image transmission. It has a non-trivial cost, and
            # at the moment Meldis can't display LCM images anyway.
            self._builder.builder().RemoveSystem(camera_publisher)

        # Add joint sliders to meshcat.
        # TODO(trowell-tri) Restoring slider values depends on the slider
        # names remaining consistent across the reload; currently the
        # JointSlider code names sliders by joint name and position suffix
        # and adds a model name suffix when those names aren't unique, e.g.
        # when the same model appears multiple times. Thus if the set of
        # models changes to cause or eliminate such a name collision then
        # slider values won't be fully restored. It would probably be better to
        # be able to configure the JointSliders to always use fully-qualified
        # names on demand, especially once the size of the control panel is
        # adjustable so that the slider names are better visible.
        self._sliders = self._builder.builder().AddNamedSystem(
            "joint_sliders",
            JointSliders(meshcat=self._meshcat, plant=self._builder.plant()),
        )

        # Connect to PyPlot.
        if self._pyplot:
            ConnectPlanarSceneGraphVisualizer(
                self._builder.builder(), self._builder.scene_graph()
            )

        self._original_package_map = copy.copy(
            self._builder.parser().package_map()
        )
        self._diagram = self._builder.Build()
        self._builder = None
        self._context = self._diagram.CreateDefaultContext()

        # We don't just test 'position' because NumPy does weird things with
        # the truth values of arrays.
        if position is not None and len(position) > 0:
            self._raise_if_invalid_positions(position)
            self._diagram.plant().SetPositions(
                self._diagram.plant().GetMyMutableContextFromRoot(
                    self._context
                ),
                position,
            )
            self._sliders.SetPositions(position)

        # Use Simulator to dispatch initialization events.
        # TODO(eric.cousineau): Simplify as part of #13776 (was #10015).
        Simulator(self._diagram).Initialize()
        # Publish draw messages with current state.
        self._diagram.ForcedPublish(self._context)

        self._check_rep(finalized=True)

    @staticmethod
    def _camera_config_to_frustum(camera: CameraConfig):
        """
        Returns a mesh as (vertices, faces) to visualize the given camera's
        frustum.
        """
        distance = camera.z_far
        width = 0.5 * camera.width * distance / camera.focal_x()
        height = 0.5 * camera.height * distance / camera.focal_y()
        vertices = np.array(
            [
                [0.0, 0.0, 0.0],
                [+width, +height, distance],
                [+width, -height, distance],
                [-width, -height, distance],
                [-width, +height, distance],
            ]
        ).T
        faces = np.array(
            [
                [0, 1, 2],
                [0, 2, 3],
                [0, 3, 4],
                [0, 4, 1],
            ]
        ).T
        return (vertices, faces)

    def _reload(self):
        """
        Re-creates the Diagram using the same sequence of calls to AddModels
        as the user performed. In effect, this will refresh the visualizer to
        show any changes the user made on disk to their models.
        """
        self._check_rep(finalized=True)
        assert self._added_models is not None

        # Clear out the old diagram.
        self._diagram = None
        self._sliders = None
        self._context = None
        self._remove_traffic_cone()

        # Populate the diagram builder again with the same packages and models.
        self._builder = RobotDiagramBuilder()
        self._builder.parser().SetAutoRenaming(True)
        self._builder.parser().package_map().AddMap(self._original_package_map)
        try:
            for kwargs in self._added_models:
                self._builder.parser().AddModels(**kwargs)
            logging.getLogger("drake").info("Reload was successful")
        except BaseException as e:
            # If there's a parsing error, show it; don't crash.
            logging.getLogger("drake").error(e)
            logging.getLogger("drake").warning(
                f"Click '{self._reload_button_name}' to try again"
            )
            # Clear the display to help indicate the failure to the user.
            self._builder = RobotDiagramBuilder()
            self._builder.parser().package_map().AddMap(
                self._original_package_map
            )
            self._add_traffic_cone()
        self._original_package_map = None

        # Finalize the rest of the systems and widgets.
        self.Finalize()

    def _render_if_necessary(self, *, loop_once):
        """This evaluates the state of the camera and plant and possibly
        triggers a rendering (up to a maximum hard-coded rate).
        """
        if not self._show_rgbd_sensor:
            return
        # When we only have a loop once, we should get one rendering.
        if not loop_once and time.time() < (self._last_camera_time + 0.0625):
            return

        self._last_camera_time = time.time()
        X_WC = self._meshcat.GetTrackedCameraPose()
        if X_WC is None:
            return

        frame = self._diagram.plant().GetFrameByName("$rgbd_sensor_offset")
        frame.SetPoseInParentFrame(
            context=self._diagram.plant().GetMyMutableContextFromRoot(
                self._context
            ),
            X_PF=X_WC,
        )
        self._diagram.GetOutputPort("preview_image").Eval(self._context)

    def Run(self, position=None, loop_once=False):
        """
        Runs the model. If Finalize() hasn't already been explicitly called
        then the object will be finalized first.

        Will iterate once and exit if `loop_once` is True, otherwise will
        loop until the user quits.

        Args:
          position: an ndarray-like list of slider positions for the model(s);
            must match the number of positions in all model(s), including the
            7 positions corresponding to any model bases that aren't welded to
            the world.
          loop_once: a flag that exits the evaluation loop after one pass.
        """
        if self._builder is not None:
            self.Finalize(position=position)
        else:
            self._check_rep(finalized=True)
            if position is not None and len(position) > 0:
                self._raise_if_invalid_positions(position)
                self._diagram.plant().SetPositions(
                    self._diagram.plant().GetMyMutableContextFromRoot(
                        self._context
                    ),
                    position,
                )
                self._sliders.SetPositions(position)
                self._diagram.ForcedPublish(self._context)

        # Everything is finally fully configured. We can open the window now.
        # TODO(jwnimmer-tri) The browser_new config knob would probably make
        # more sense as an argument to Run() vs an argument to our constructor.
        if self._browser_new:
            self._browser_new = False
            url_params = ""
            if self._show_rgbd_sensor:
                url_params = "?tracked_camera=on"
            url = self._meshcat.web_url() + url_params
            _webbrowser_open(url=url, new=True)
        elif self._show_rgbd_sensor:
            logging.getLogger("drake").info(
                "You've requested to show the RGBD Sensor. To control the "
                "sensor position, make sure you open one browser to the "
                "following url:\n\n"
                f"\t{self._meshcat.web_url()}?tracked_camera=on\n"
            )

        # Wait for the user to cancel us.
        stop_button_name = "Stop Running"
        if not loop_once:
            logging.getLogger("drake").info(
                f"Click '{stop_button_name}' or press Esc to quit"
            )

        try:
            self._meshcat.AddButton(stop_button_name, "Escape")

            def has_clicks(button_name):
                if not button_name:
                    return False
                return self._meshcat.GetButtonClicks(button_name) > 0

            while True:
                self._render_if_necessary(loop_once=loop_once)
                time.sleep(1 / 32.0)
                if has_clicks(self._reload_button_name):
                    self._meshcat.DeleteButton(stop_button_name)
                    slider_values = self._get_slider_values()
                    self._reload()
                    self._set_slider_values(slider_values)
                    self._meshcat.AddButton(stop_button_name, "Escape")
                q = self._sliders.get_output_port().Eval(
                    self._sliders.GetMyContextFromRoot(self._context)
                )
                self._diagram.plant().SetPositions(
                    self._diagram.plant().GetMyMutableContextFromRoot(
                        self._context
                    ),
                    q,
                )
                self._diagram.ForcedPublish(self._context)
                if loop_once or has_clicks(stop_button_name):
                    break
        except KeyboardInterrupt:
            pass

        self._meshcat.DeleteButton(stop_button_name)
        if self._reload_button_name is not None:
            self._meshcat.DeleteButton(self._reload_button_name)
            self._reload_button_name = None

    def _raise_if_invalid_positions(self, position):
        """
        Validate the position argument.

        Raises:
          ValueError: if the length of the position list does not match
        the number of positions in the plant.
        """
        assert self._diagram is not None
        actual = len(position)
        expected = self._diagram.plant().num_positions()
        if actual != expected:
            raise ValueError(
                f"Number of passed positions ({actual}) does not match the "
                f"number in the model ({expected})."
            )

    def _get_slider_values(self):
        """Returns a map of slider names to current values."""
        return {
            name: self._meshcat.GetSliderValue(name)
            for name in self._meshcat.GetSliderNames()
        }

    def _set_slider_values(self, slider_values):
        """
        Sets current sliders to the values found in the slider_values dict.

        Current sliders not in the passed map -- or values in the map but
        which not longer exist in the GUI -- are ignored.
        """
        current_names = self._meshcat.GetSliderNames()
        for old_name, old_value in slider_values.items():
            if old_name in current_names:
                self._meshcat.SetSliderValue(old_name, old_value)

    def _add_traffic_cone(self):
        """Adds a traffic cone to the scene, indicating a parsing error."""
        base_width = 0.4
        base_thickness = 0.01
        base = Box(base_width, base_width, base_thickness)
        cone_height = 0.6
        cone_radius = 0.75 * (base_width / 2)
        cone = MeshcatCone(cone_height, cone_radius, cone_radius)

        path = "/PARSE_ERROR"
        orange = Rgba(1.0, 0.33, 0)
        self._meshcat.SetObject(path=f"{path}/base", shape=base, rgba=orange)
        self._meshcat.SetObject(path=f"{path}/cone", shape=cone, rgba=orange)
        self._meshcat.SetTransform(
            f"{path}/base", RigidTransform([0, 0, base_thickness * 0.5])
        )
        self._meshcat.SetTransform(
            f"{path}/cone",
            RigidTransform(
                RotationMatrix.MakeYRotation(np.pi),
                [0, 0, base_thickness + cone_height],
            ),
        )

    def _remove_traffic_cone(self):
        """Removes the traffic cone from the scene."""
        self._meshcat.Delete("/PARSE_ERROR")
